/*
 * Copyright (c) 2014, 2020 Oracle and/or its affiliates. All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0, which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the
 * Eclipse Public License v. 2.0 are satisfied: GNU General Public License,
 * version 2 with the GNU Classpath Exception, which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 */

package org.glassfish.jersey.tests.e2e.client.connector;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import jakarta.ws.rs.HeaderParam;
import jakarta.ws.rs.POST;
import jakarta.ws.rs.Path;
import jakarta.ws.rs.Priorities;
import jakarta.ws.rs.WebApplicationException;
import jakarta.ws.rs.client.AsyncInvoker;
import jakarta.ws.rs.client.ClientRequestContext;
import jakarta.ws.rs.client.ClientRequestFilter;
import jakarta.ws.rs.client.Entity;
import jakarta.ws.rs.client.Invocation;
import jakarta.ws.rs.core.Application;
import jakarta.ws.rs.core.MediaType;
import jakarta.ws.rs.core.MultivaluedMap;
import jakarta.ws.rs.core.Response;
import jakarta.ws.rs.ext.MessageBodyWriter;
import jakarta.ws.rs.ext.WriterInterceptor;
import jakarta.ws.rs.ext.WriterInterceptorContext;

import javax.annotation.Priority;

import org.glassfish.jersey.apache.connector.ApacheConnectorProvider;
import org.glassfish.jersey.client.ClientConfig;
import org.glassfish.jersey.client.HttpUrlConnectorProvider;
import org.glassfish.jersey.client.spi.ConnectorProvider;
import org.glassfish.jersey.logging.LoggingFeature;
import org.glassfish.jersey.grizzly.connector.GrizzlyConnectorProvider;
import org.glassfish.jersey.jetty.connector.JettyConnectorProvider;
import org.glassfish.jersey.server.ResourceConfig;
import org.glassfish.jersey.test.JerseyTest;
import org.glassfish.jersey.test.TestProperties;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import static org.hamcrest.Matchers.containsString;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertThat;

/**
 * JERSEY-2206 reproducer
 *
 * @author Libor Kramolis
 */
@RunWith(Parameterized.class)
public class RequestHeaderModificationsTest extends JerseyTest {

    private static final Logger LOGGER = Logger.getLogger(RequestHeaderModificationsTest.class.getName());
    private static final boolean GZIP = false; // change to true when JERSEY-2341 fixed
    private static final boolean DUMP_ENTITY = false; // I have troubles to dump entity with async jetty!

    private static final String QUESTION = "QUESTION";
    private static final String ANSWER = "ANSWER";
    private static final String REQUEST_HEADER_NAME_CLIENT = "Client-Prop";
    private static final String REQUEST_HEADER_VALUE_CLIENT = "Client-Value";
    private static final String REQUEST_HEADER_NAME_FILTER = "Filter-Prop";
    private static final String REQUEST_HEADER_VALUE_FILTER = "Filter-Value";
    private static final String REQUEST_HEADER_NAME_INTERCEPTOR = "Iceptor-Prop";
    private static final String REQUEST_HEADER_VALUE_INTERCEPTOR = "Iceptor-Value";
    private static final String REQUEST_HEADER_NAME_MBW = "Mbw-Prop";
    private static final String REQUEST_HEADER_VALUE_MBW = "Mbw-Value";
    private static final String REQUEST_HEADER_MODIFICATION_SUPPORTED = "modificationSupported";
    private static final String PATH = "/resource";

    @Parameterized.Parameters(name = "{index}: {0} / modificationSupported= {1} / addHeader= {2}")
    public static List<Object[]> testData() {
        return Arrays.asList(new Object[][] {
                {HttpUrlConnectorProvider.class, true, false},
                {GrizzlyConnectorProvider.class, false, false}, // change to true when JERSEY-2341 fixed
                {JettyConnectorProvider.class, false, false}, // change to true when JERSEY-2341 fixed
                {ApacheConnectorProvider.class, false, false}, // change to true when JERSEY-2341 fixed
                {HttpUrlConnectorProvider.class, true, true},
                {GrizzlyConnectorProvider.class, false, true}, // change to true when JERSEY-2341 fixed
                {JettyConnectorProvider.class, false, true}, // change to true when JERSEY-2341 fixed
                {ApacheConnectorProvider.class, false, true}, // change to true when JERSEY-2341 fixed
        });
    }

    private final ConnectorProvider connectorProvider;
    private final boolean modificationSupported; // remove when JERSEY-2341 fixed
    private final boolean addHeader;

    public RequestHeaderModificationsTest(Class<? extends ConnectorProvider> connectorProviderClass,
                                          boolean modificationSupported, boolean addHeader)
            throws IllegalAccessException, InstantiationException {
        this.connectorProvider = connectorProviderClass.newInstance();
        this.modificationSupported = modificationSupported;
        this.addHeader = addHeader;
    }

    @Override
    protected Application configure() {
        set(TestProperties.RECORD_LOG_LEVEL, Level.WARNING.intValue());

        enable(TestProperties.LOG_TRAFFIC);
        if (DUMP_ENTITY) {
            enable(TestProperties.DUMP_ENTITY);
        }
        return new ResourceConfig(TestResource.class).register(new LoggingFeature(LOGGER, LoggingFeature.Verbosity.HEADERS_ONLY));
    }

    @Override
    protected void configureClient(ClientConfig clientConfig) {
        clientConfig.register(MyClientRequestFilter.class);
        clientConfig.register(new MyWriterInterceptor(addHeader));
        clientConfig.register(new MyMessageBodyWriter(addHeader));
        clientConfig.connectorProvider(connectorProvider);
    }

    @Test
    public void testWarningLogged() throws Exception {
        Response response = requestBuilder().post(requestEntity());
        assertResponse(response);
    }

    @Test
    public void testWarningLoggedAsync() throws Exception {
        AsyncInvoker asyncInvoker = requestBuilder().async();
        Future<Response> responseFuture = asyncInvoker.post(requestEntity());
        Response response = responseFuture.get();
        assertResponse(response);
    }

    private Invocation.Builder requestBuilder() {
        return target(PATH)
                .request()
                .header(REQUEST_HEADER_NAME_CLIENT, REQUEST_HEADER_VALUE_CLIENT)
                .header(REQUEST_HEADER_MODIFICATION_SUPPORTED, modificationSupported && addHeader)
                .header("hello", "double").header("hello", "value");
    }

    private Entity<MyEntity> requestEntity() {
        return Entity.text(new MyEntity(QUESTION));
    }

    private void assertResponse(Response response) {
        if (!modificationSupported) {
            final String UNSENT_HEADER_CHANGES = "Unsent header changes";
            LogRecord logRecord = findLogRecord(UNSENT_HEADER_CHANGES);
            if (addHeader) {
                assertNotNull("Missing LogRecord for message '" + UNSENT_HEADER_CHANGES + "'.", logRecord);
                assertThat(logRecord.getMessage(), containsString(REQUEST_HEADER_NAME_INTERCEPTOR));
                assertThat(logRecord.getMessage(), containsString(REQUEST_HEADER_NAME_MBW));
            } else {
                assertNull("Unexpected LogRecord for message '" + UNSENT_HEADER_CHANGES + "'.", logRecord);
            }
        }

        assertEquals(200, response.getStatus());
        assertEquals(ANSWER, response.readEntity(String.class));
    }

    private LogRecord findLogRecord(String messageContains) {
        for (final LogRecord record : getLoggedRecords()) {
            if (record.getMessage().contains(messageContains)) {
                return record;
            }
        }
        return null;
    }

    @Path(PATH)
    public static class TestResource {

        @POST
        public String handle(InputStream questionStream,
                             @HeaderParam(REQUEST_HEADER_NAME_CLIENT) String client,
                             @HeaderParam(REQUEST_HEADER_NAME_FILTER) String filter,
                             @HeaderParam(REQUEST_HEADER_NAME_INTERCEPTOR) String interceptor,
                             @HeaderParam(REQUEST_HEADER_NAME_MBW) String mbw,
                             @HeaderParam(REQUEST_HEADER_MODIFICATION_SUPPORTED) boolean modificationSupported)
                throws IOException {
            assertEquals(REQUEST_HEADER_VALUE_CLIENT, client);
            assertEquals(REQUEST_HEADER_VALUE_FILTER, filter);
            if (modificationSupported) {
                assertEquals(REQUEST_HEADER_VALUE_INTERCEPTOR, interceptor);
                assertEquals(REQUEST_HEADER_VALUE_MBW, mbw);
            }
            assertEquals(QUESTION, new Scanner(GZIP ? new GZIPInputStream(questionStream) : questionStream).nextLine());
            return ANSWER;
        }
    }

    public static class MyWriterInterceptor implements WriterInterceptor {

        private final boolean addHeader;

        public MyWriterInterceptor(boolean addHeader) {
            this.addHeader = addHeader;
        }

        @Override
        public void aroundWriteTo(WriterInterceptorContext context) throws IOException, WebApplicationException {
            if (addHeader) {
                context.getHeaders().add(REQUEST_HEADER_NAME_INTERCEPTOR, REQUEST_HEADER_VALUE_INTERCEPTOR);
            }
            if (GZIP) {
                context.setOutputStream(new GZIPOutputStream(context.getOutputStream()));
            }
            context.proceed();
        }
    }

    public static class MyClientRequestFilter implements ClientRequestFilter {

        @Override
        public void filter(ClientRequestContext requestContext) throws IOException {
            requestContext.getHeaders().add(REQUEST_HEADER_NAME_FILTER, REQUEST_HEADER_VALUE_FILTER);
        }
    }

    @Priority(Priorities.ENTITY_CODER)
    public static class MyMessageBodyWriter implements MessageBodyWriter<MyEntity> {

        private final boolean addHeader;

        public MyMessageBodyWriter(boolean addHeader) {
            this.addHeader = addHeader;
        }

        @Override
        public boolean isWriteable(Class type, Type genericType, Annotation[] annotations, MediaType mediaType) {
            return true;
        }

        @Override
        public long getSize(MyEntity o, Class type, Type genericType, Annotation[] annotations, MediaType mediaType) {
            return -1; //ignored
        }

        @Override
        public void writeTo(MyEntity o, Class type, Type genericType, Annotation[] annotations, MediaType mediaType,
                            MultivaluedMap httpHeaders, OutputStream entityStream)
                throws IOException, WebApplicationException {
            if (addHeader) {
                httpHeaders.add(REQUEST_HEADER_NAME_MBW, REQUEST_HEADER_VALUE_MBW);
            }
            entityStream.write(o.getValue().getBytes());
        }
    }

    public static class MyEntity {

        private String value;

        public MyEntity() {
        }

        public MyEntity(String value) {
            this.value = value;
        }

        public String getValue() {
            return value;
        }

        public void setValue(String value) {
            this.value = value;
        }
    }

}
